/*
 *  GreedyRWA.java
 *
 *  Copyright (C) 1997 Esa Hyyti
 *
 *  Greedy RWA heuristic.
 *
 */

import java.awt.*;
import java.lang.Math;
import java.util.Vector;

/*****************************************************/

class GreedyRWA
{
    static boolean DEBUG = false;

    /*
     * Variables
     */
    BinaryMatrix cm;     /* problem           */
    Vector       routes; /* solution, routes  */
    int          colors; /* and nr. of colors */
    
    /*
     * Constructor
     */
    public GreedyRWA( BinaryMatrix lm ) 
    {
        cm     = lm;
        routes = null;
        colors = -1;
    }

    public int    getNumberOfColors( )  { return colors; }
    public Vector getConfiguration( )   { return routes; }

    /*
     *  Picks the best routes between each pair, local random search
     *  Returns RWA (Routes stored in vector)
     */
    public boolean runRWA( ) 
    {
        Vector allroutes = new Vector();
        BinaryMatrix lm;
        int i,j;
        int best;
        
        /* generate all the shortest paths between each pair */
        for( i=0; i<cm.np; i++)
            for ( j=i+1; j<cm.np; j++)
            {
                Routes  ijroutes  = new Routes( cm, i, j );
                if ( ijroutes.status() == false )
                {
                    System.out.println( "Can't fully connect the graph." );
                    return false;
                }
                allroutes.addElement( ijroutes );
            }

        /* ok, allroutes has now all the route sets. At start pick the first ones */
        int selection[]     = new int[ allroutes.size() ];
        int choices[]       = new int[ allroutes.size() ];
        boolean can_iterate = false;

        for ( i=0; i<allroutes.size(); i++ )
        {
            selection[i] = 0;
            choices[i]   = ((Routes)allroutes.elementAt(i)).routes.size();
            if ( choices[i]>1 )
                can_iterate = true;
        }

        /* Assign the wavelengths to chosen paths */
        lm               = getWAMatrix( allroutes, selection );
        NodeColoring col = new NodeColoring( lm );
        best = col.num;
        
        /* time to iterate */
        if ( can_iterate )
        {
            for ( int iter=0; iter<30; iter++ )
            {
                int nr;
                int or;

                do 
                    i = (int)(allroutes.size() * Math.random());
                while ( i>=allroutes.size() || choices[i]<2 );
                do 
                    nr = (int)(choices[i] * Math.random());
                while ( nr>=choices[i] || selection[i]==nr );

                or = selection[i];
                selection[i] = nr;

                lm = getWAMatrix( allroutes, selection );
                col = new NodeColoring( lm );

                if ( col.num > best )
                    selection[i] = or;
                else
                    best = col.num;
            }
        }
        if ( DEBUG )
            System.out.println( "Best:" + best + " colors." );

        lm = getWAMatrix( allroutes, selection );
        col = new NodeColoring( lm );
        col.print();

        /* ok, selection now has the best set of routes */
        Vector best_routes = new Vector();
        for (i=0; i<allroutes.size(); i++ )
        {
            Route r1 = ( (Routes)allroutes.elementAt(i) ).getRoute( selection[i] );
            r1.color = col.color[i];
            best_routes.addElement( r1 );
        }

        /* store the results */
        colors = best;
        routes = best_routes;
        return true;
    }

    /*****************************************************************/

    /*
     *  Returns true if routes have a common link.
     *  (very unefficient implementation)
     */
    boolean commonLink( Route a, Route b ) 
    {
        for ( int i=0; i<a.d; i++ )
        {
            for( int j=0; j<b.d; j++ )
            {
                if ( a.node[i]==b.node[j] )
                {
                    if ( i>0 )
                    {
                        if ( j>0     && a.node[i-1] == b.node[j-1] )  return true;
                        if ( j+1<b.d && a.node[i-1] == b.node[j+1] )  return true;
                    }
                    if ( i+1<a.d )
                    {
                        if ( j>0     && a.node[i+1] == b.node[j-1] )  return true;
                        if ( j+1<b.d && a.node[i+1] == b.node[j+1] )  return true;
                    }
                }
            }
        }
        return false;
    }


    /*
     *  Returns a wavelength assignment matrix based on selected routes
     */
    BinaryMatrix getWAMatrix( Vector allroutes, int selection[] ) 
    {
        int i,j;
        BinaryMatrix lm;
        Routes routes;
        Route r1,r2;

        lm = new BinaryMatrix( allroutes.size() );
        
        for ( i=0; i<allroutes.size(); i++ )
        {
            routes = (Routes)allroutes.elementAt(i);
            r1     = routes.getRoute( selection[i] );

            for ( j=i+1; j<allroutes.size(); j++ )
            {
                routes = (Routes)allroutes.elementAt(j);
                r2     = routes.getRoute( selection[j] );
                if ( commonLink( r1, r2 ) )
                {
                    lm.m[i][j] = true;
                    lm.m[j][i] = true;
                }
            }
        }
        return lm;
    }
}
