/*
 *  WdmDemo.java
 *
 *  Copyright (C) 1997 Esa Hyyti
 *
 *  A simple java applet demonstrating the static RWA in optical networks
 *
 */

import java.awt.*;
import java.applet.*;
import java.lang.Math;
import java.util.Vector;

/*****************************************************/

class Config
{
    static final int MODE_NONE    = 0;
    static final int MODE_ADD     = 1;
    static final int MODE_MOVE    = 2;
    static final int MODE_CONNECT = 3;
    static final int MODE_REMOVE  = 4;
    static final int MODE_MAX     = 5;

    static final String MODE_NAME[] =
    {
        "<none>",
        "add nodes",
        "move nodes",
        "connect nodes",
        "remove nodes",
        "<none>"
    };

    static final int SINGLE_LINE  = 1;
    static final int DOUBLE_LINE  = 2;
    
    public Config( )
    {
        mode     = MODE_ADD;
        drawMode = SINGLE_LINE;
    }

    int mode;
    int drawMode;
    
}

/*****************************************************/

public class WdmDemo extends Applet
{
    public void init() 
    {
        DrawPanel    dp = new DrawPanel();
        DrawControls dc = new DrawControls(dp);
        
        setLayout(new BorderLayout());

        add("Center", dp );
        add("South",  dc );
    }

    public boolean handleEvent(Event e) 
    {
        switch (e.id) {
            case Event.WINDOW_DESTROY:
                System.exit(0);
                return true;
            default:
                return false;
        }
    }
    
    public static void main(String args[]) 
    {
        Frame   f       = new Frame("WdmDemo");
        WdmDemo WdmDemo = new WdmDemo();

        WdmDemo.init();
        WdmDemo.start();
        
        f.add("Center", WdmDemo);
        f.resize(500, 300);
        f.show();
    }
}

/*****************************************************/

class DrawPanel extends Panel 
{
    int    mode      = Config.MODE_NONE;     /* current mode, 'triggered' */
    
    /* variables */
    Config config    = new Config();
    Vector nodes     = new Vector();
    Node   node_from = null;
    Vector routes    = null;
    int    colors    = 0;
    int    curr_x, curr_y;
    
    public DrawPanel() 
    {
        setBackground(Color.lightGray);
        initStartup( );
    }
    
    public Node getNode( int x, int y ) 
    {
        Node closest = null;
        int distance = 0;
        
        int np = nodes.size();
        for (int i=0; i < np; i++) {
            Node n = (Node)nodes.elementAt(i);
            int d  = (n.r.x - x)*(n.r.x - x) + (n.r.y - y)*(n.r.y - y);
            if ( d < distance || closest==null )
            {
                closest=n;
                distance=d;
            }
        }

        if ( distance < 400 )
            return closest;
        else
            return null;
    }

    /*
     *  Returns a matrix representing the physical links
     */
    BinaryMatrix getLinkMatrix( )
    {
        int i,j;
        int n  = nodes.size();

        if ( n < 1 )
            return new BinaryMatrix();

        BinaryMatrix lm = new BinaryMatrix( n );

        for (i=0; i < n; i++) 
        {
            Node n1 = (Node)nodes.elementAt(i);
            for (j=0; j<n; j++ ) 
            {
                if ( i!=j)
                {
                    Node n2 = (Node)nodes.elementAt(j);
                    if ( n1.neighbours.indexOf( n2 ) != -1 )
                        lm.m[i][j] = true;
                }
            }
        }

        return lm;
    }

    public void handleRWA() 
    {
        GreedyRWA rwa = new GreedyRWA( getLinkMatrix( ) );

        if ( rwa.runRWA() )
        {
            colors = rwa.getNumberOfColors( );
            routes = rwa.getConfiguration( );
            repaint();
        }
    }
    
    public void clearNodes() 
    {
        routes = null;
        nodes.removeAllElements();
        repaint();
    }

    // handlers
    private void start_connect( Event e )
    {
        mode      = Config.MODE_CONNECT;
        node_from = getNode( e.x, e.y );
        routes    = null;
    }

    private void finish_connect( Event e )
    {
        Node n;
        
        if ( node_from == null )
            return;
        
        if ( ( n = getNode( e.x, e.y ) ) != null )
        {
            if ( !n.neighbours.contains( node_from ) && n != node_from )
            {
                n.neighbours.addElement( node_from );
                node_from.neighbours.addElement( n );
                routes = null;
            }
        }
        node_from = null;
        mode      = Config.MODE_NONE;
        repaint();
    }
    
    private void remove_node( Event e )        
    {
        Node n = getNode( e.x, e.y );
        if ( n != null )
        {
            int np = nodes.size();
            for ( int i=0; i < np; i++) {
                Node n2 = (Node)nodes.elementAt(i);
                n2.neighbours.removeElement( n );
            }
            nodes.removeElement( n );
            routes = null;
            repaint();
        }
    }
    
    private void add_node( Event e )
    {
        nodes.addElement(new Node(e.x, e.y));
        routes = null;
        repaint();
    }

    public boolean handleEvent(Event e) 
    {
        Node n;
        int i;
        
        switch (e.id) {
            case Event.MOUSE_DOWN:
                if ( mode != Config.MODE_NONE )
                    break;
                if ( e.modifiers ==  e.ALT_MASK )   // middle button -> connect
                    start_connect( e );
                else if ( e.modifiers == e.META_MASK )  // right button --> delete
                    remove_node( e );             
                else
                {
                    switch( config.mode )
                    {
                        case Config.MODE_ADD:
                            add_node( e );
                            break;
                        case Config.MODE_MOVE:
                            if ( ( node_from = getNode( e.x, e.y ) ) != null )
                                mode = Config.MODE_MOVE;
                            break;
                        case Config.MODE_REMOVE:
                            remove_node( e );
                            break;
                        case Config.MODE_CONNECT:
                            start_connect( e );
                            break;
                    }
                }
                
                return true;

            case Event.MOUSE_UP:
                if ( mode == Config.MODE_CONNECT ) // e.modifiers == e.ALT_MASK && 
                    finish_connect( e );
                mode = Config.MODE_NONE;
                return true;

            case Event.MOUSE_DRAG:
                if ( mode == Config.MODE_CONNECT )
                {
                    if ( node_from == null )
                        mode = Config.MODE_NONE;
                    else
                    {
                        curr_x = e.x; curr_y = e.y;
                        repaint();
                    }
                }
                else if ( mode == Config.MODE_MOVE )
                {
                    if ( node_from == null )
                        mode = Config.MODE_NONE;
                    else
                    {
                        node_from.r.x = e.x;
                        node_from.r.y = e.y;
                        repaint();
                    }
                }
                return true;

            case Event.WINDOW_DESTROY:
                System.exit(0);
                return true;
        }

        return false;
    }
    
    double distance( Point p1, Point p2 )
    {
        return Math.sqrt( (p1.x - p2.x)*(p1.x - p2.x) +
                          (p1.y - p2.y)*(p1.y - p2.y) );
    }
    
    void fixDistance( Point p1, Point p2, Point center, int d )
    {
        Point v       = new Point( p2.x, p2.y );
        double scale  = 1;
        
        if ( distance( p1, center ) < d )
            return;

        v.translate( -p1.x, -p1.y );
        while( distance( p2, center ) < d )
        {
            scale = scale - 0.01;
            p2.move( p1.x, p1.y );
            p2.translate( (int)(scale*v.x), (int)(scale*v.y) );
        }
    }

    /*************************************************************/

    /*
     *  Graphic routines
     */
    void drawRound( Graphics g, Point p1, Point p2, Point center )
    {
        int r = (int)( distance( p1, center ));
        int theta1, theta2;

        theta1 = (int)( 180*Math.atan2( -p1.y+center.y, p1.x-center.x )/Math.PI );
        theta2 = (int)( 180*Math.atan2( -p2.y+center.y, p2.x-center.x )/Math.PI );
        theta2 = theta2 - theta1;
        if      ( theta2 >  180 )            theta2 -= 360;
        else if ( theta2 < -180 )            theta2 += 360;

        g.drawArc( center.x-r, center.y-r, 2*r, 2*r, theta1, theta2 );
        if ( config.drawMode == Config.DOUBLE_LINE )
            g.drawArc( center.x-r-1, center.y-r-1, 2*r+2, 2*r+2, theta1, theta2 );
    }

    void drawLink( Graphics g, Point p1, Point p2 )
    {
        g.drawLine( p1.x,   p1.y,   p2.x,   p2.y   );
        
        if ( config.drawMode == Config.DOUBLE_LINE )
        {
            g.drawLine( p1.x+1, p1.y,   p2.x+1, p2.y   ); 
            g.drawLine( p1.x,   p1.y+1, p2.x,   p2.y+1 );
            g.drawLine( p1.x+1, p1.y+1, p2.x+1, p2.y+1 );
        }
    }
    
    void drawColors( Graphics g )
    {
        Route  r;
        Point  p1, p2, p3, p4, delta;
        int    i,j;
        int    np    = nodes.size();  
        int    width = 4;
        Node   n1, n2;

        for ( i=0; i<np; i++ )
        {
            n1          = (Node)nodes.elementAt( i );
            n1.circlers = 0;
        }

        p1    = new Point(0,0);
        p2    = new Point(0,0);
        p3    = new Point(0,0);
        p4    = new Point(0,0);
        delta = new Point(0,0);
            
        for ( i=0; i<routes.size(); i++ )
        {
            r  = (Route)routes.elementAt(i);
            n2 = (Node)nodes.elementAt(r.node[0]);
            g.setColor( ColorConversion.intToColor(r.color) );
            
            for (j=1; j<r.d; j++ )
            {
                double length, nx, ny;
                    
                n1 = n2;
                n2 = (Node)nodes.elementAt(r.node[j]);

                /* calculate normal */
                length = distance( n1.r, n2.r );
                if ( length <= 0.01 )
                    length = 1;
                nx = width*(n2.r.y - n1.r.y)/length;
                ny = width*(n1.r.x - n2.r.x)/length;
                if (nx<0 || ( nx == 0 && ny<0 ) )
                {
                    nx = -nx;
                    ny = -ny;
                }
                delta.x = (int)((r.color/2)*nx);
                delta.y = (int)((r.color/2)*ny);
                if ( r.color % 2 == 0 )
                {
                    delta.x = -delta.x;
                    delta.y = -delta.y;
                }
                    
                p3.move( n1.r.x, n1.r.y );  p3.translate( delta.x, delta.y );
                p4.move( n2.r.x, n2.r.y );  p4.translate( delta.x, delta.y );

                if ( j!=1 )   // do nothing on first hop
                {
                    int cr = (int)( n1.radius + (n1.circlers+2)*width );
                    n1.circlers++;
                    fixDistance( p1, p2, n1.r, cr );
                    fixDistance( p4, p3, n1.r, cr );
                    drawLink(  g, p1, p2 );
                    drawRound( g, p2, p3, n1.r );
                }
                p1.move( p3.x, p3.y );
                p2.move( p4.x, p4.y );
            }
            drawLink( g, p1, p2 );
        }
    }
    
    public void paint(Graphics g) 
    {
        int np, i, j;
        Node n1,n2;
        
        np     = nodes.size();     // clear counter 
        
        if ( routes != null && colors < 9 )
            drawColors( g );
        else
        {   /* Draw links then */
            int nn;
            
            g.setColor(Color.black);       /* TWICE! */
            for ( i=0; i < np; i++) {
                n1 = (Node)nodes.elementAt(i);
                nn = n1.neighbours.size();
                for ( j=0; j<nn; j++ )
                {
                    n2 = (Node)n1.neighbours.elementAt(j);
                    g.drawLine( n1.r.x, n1.r.y, n2.r.x, n2.r.y );
                }
            }
        }

        /* Draw Nodes */
        for ( i=0; i < np; i++)            /* then nodes */
        {
            n1 = (Node)nodes.elementAt(i);
            n1.paint(g);
        }

        /* draw line if connecting */
        if ( mode == Config.MODE_CONNECT && node_from != null )
        {
            g.setColor(Color.black );
            g.drawLine( node_from.r.x, node_from.r.y, curr_x, curr_y );
        }
    }

    /****************************************************************/
    
    /* default start when applet starts */
    void initStartup( ) 
    {
        int  dx,dy, x0, y0;
        Node n1, n2, n3, n4, n5, n6;
        
        dx = 60;
        dy = (int)( Math.sqrt(3) * dx );
        
        x0 =  90;
        y0 = -40;
        
        n1 = new Node( x0+5*dx, y0+2*dy );
        n2 = new Node( x0+4*dx, y0+3*dy );
        n3 = new Node( x0+2*dx, y0+3*dy );
        n4 = new Node( x0+1*dx, y0+2*dy );
        n5 = new Node( x0+2*dx, y0+1*dy );
        n6 = new Node( x0+4*dx, y0+1*dy );
        
        nodes.addElement( n1 );
        nodes.addElement( n2 );
        nodes.addElement( n3 );
        nodes.addElement( n4 );
        nodes.addElement( n5 );
        nodes.addElement( n6 );

        n1.neighbours.addElement( n2 );
        n1.neighbours.addElement( n6 );
        n2.neighbours.addElement( n1 );
        n2.neighbours.addElement( n3 );
        n3.neighbours.addElement( n2 );
        n3.neighbours.addElement( n4 );
        n4.neighbours.addElement( n3 );
        n4.neighbours.addElement( n5 );
        n5.neighbours.addElement( n4 );
        n5.neighbours.addElement( n6 );
        n6.neighbours.addElement( n5 );
        n6.neighbours.addElement( n1 );
    }

}

/*****************************************************/

class DrawControls extends Panel 
{
    DrawPanel target;
    TextField status;

    public DrawControls(DrawPanel target) 
    {
        this.target = target;
        setLayout(new FlowLayout());
        setBackground(Color.lightGray);
        target.setForeground(Color.red);
        //CheckboxGroup group = new CheckboxGroup();
        //Checkbox      b;
        Button        button;
        Choice        choice  = new Choice();

        for( int i=1; i<Config.MODE_MAX; i++ )
            choice.addItem( Config.MODE_NAME[i] );
        choice.setBackground(Color.lightGray);
        add( choice );
        
        add( button = new Button( "Assign" ) );
        add( button = new Button( "Reset"  ) );
        add( status = new TextField( "A simple WDM-demonstrator is ready.", 40 ) );
        status.setEditable( false );
    }

    public void paint(Graphics g) 
    {
        Rectangle r = bounds();

        g.setColor(Color.lightGray);
        g.draw3DRect(0, 0, r.width, r.height, false);
    }

    private void setDrawingMode( String str )
    {
        for( int i=1; i<Config.MODE_MAX; i++ )
        {
            if ( str.equals( Config.MODE_NAME[i] ) )
            {
                target.config.mode = i;
                return;
            }
        }
    }
    
    public boolean action(Event e, Object arg) 
    {
        /*
          if (e.target instanceof Checkbox) 
          { 
          target.setForeground(((Component)e.target).getBackground());
          }   
        */
        if (e.target instanceof Choice) 
        {
            String str = (String)arg;
            setDrawingMode( str );
	} 
        else if (e.target instanceof Button) 
        {
            String choice = (String)arg;
	    if      ( choice.equals( "Reset"  ) )
            {
                target.clearNodes();
                status.setText( "Network erased" );
            }
            else if ( choice.equals( "Assign" ) )
            {
                String s;
                
                target.handleRWA();

                if ( target.colors > 0 )
                    s = "Configuration using " + String.valueOf( target.colors ) + " wavelength channels found";
                else
                    s = "Network can't be fully connected";
                status.setText( s );
            }
        }
        return true;
    }
}

